REM
REM  ########################################################
REM  # Digger Dwarf - A console-based gopher client         #
REM  # Author: Mateusz Viste                                #
REM  # License: GNU/GPL v3                                  #
REM  # Compiled using: FreeBASIC v0.20 + chiSock library    #
REM  # http://www.viste-family.net/mateusz/software/ddwarf/ #
REM  ########################################################
REM
DECLARE FUNCTION CenterString(WorkString AS STRING, xWidth AS INTEGER) AS STRING
DECLARE SUB GetScreenSize()
DECLARE SUB LoadSettings()
DECLARE SUB ShowHelp()
DECLARE SUB RefreshScreen()
DECLARE SUB AddHistory(UrlToAdd AS STRING, SearchQueryToAdd AS STRING = "")
DECLARE SUB HistoryBack()
DECLARE SUB QuitProgram(QuitMessage AS STRING = "")
DECLARE FUNCTION GetUserString(xposition AS INTEGER, yposition AS INTEGER, fieldwidth AS INTEGER, FgColor AS INTEGER, BgColor AS INTEGER, PreFilledString AS STRING) AS STRING
DECLARE FUNCTION TypeNewURL(ActualUrl AS STRING) AS STRING
DECLARE SUB KeybFlush()
DECLARE FUNCTION RepeatPrint(RepeatString AS STRING, RepeatAction AS INTEGER) AS STRING
DECLARE FUNCTION IsGopherUrlValid(UrlToValidate AS STRING) AS BYTE
DECLARE SUB ClearContentSpace()
DECLARE SUB DrawTUI()
DECLARE FUNCTION AskForSearchRequest() AS STRING
DECLARE FUNCTION SelectFileToSave() AS STRING
DECLARE FUNCTION DumpLoadedFileTo(FileToDumpTo AS STRING) AS BYTE
DECLARE SUB SetupMenu()
DECLARE SUB UpdateStatusBar(StatusInformation AS STRING)

CONST pVer AS STRING = "0.82"
CONST pDate AS STRING = "2009"

REDIM SHARED FileBuffer(1 TO 1024) AS UBYTE
DIM SHARED UrlHistory(0 TO 255) AS STRING
DIM SHARED FileSize AS INTEGER
DIM SHARED AS INTEGER xSize, ySize, OrgCursorPosX, OrgCursorPosY
DIM SHARED AS STRING CurURL, NewUrl, SaveTo, DirDelimiter, SearchQuery
DIM SHARED AS BYTE DoNotReload

TYPE ColorStyle
  TitleBarFg AS BYTE
  TitleBarBg AS BYTE
  StatusBarFg AS BYTE
  StatusBarBg AS BYTE
  UrlBarFg AS BYTE
  UrlBarBrightFg AS BYTE
  UrlBarBg AS BYTE
  BorderFg AS BYTE
  BorderBg AS BYTE
  MsgWindowFg AS BYTE
  MsgWindowBg AS BYTE
  MsgWindowBorder AS BYTE
  DocumentFg AS BYTE
  DocumentBg AS BYTE
  GopherSelectedLinkFg AS BYTE
  GopherSelectedLinkBg AS BYTE
  GopherItemType AS BYTE
  GopherSelectedItemType AS BYTE
  GopherLink AS BYTE
  GopherBackground AS BYTE
  GopherInfo AS BYTE
  GopherError AS BYTE
  SaveFileDialogPathFg AS BYTE
  SaveFileDialogPathBg AS BYTE
  SearchTermDialogFg AS BYTE
  SearchTermDialogBg AS BYTE
END TYPE
DIM SHARED ColorScheme AS ColorStyle
DIM SHARED BoxChar(1 TO 6) AS STRING
DIM SHARED SymbolChar(1 TO 3) AS STRING
#IFDEF __FB_LINUX__    ' UTF8 encoding
  BoxChar(1) = "╔" : BoxChar(2) = "═" : BoxChar(3) = "╗"
  BoxChar(4) = "║" : BoxChar(5) = "╚" : BoxCHar(6) = "╝"
  SymbolChar(1) = "↑" : SymbolChar(2) = "░" : SymbolChar(3) = "↓"
  DirDelimiter = "/"
#ENDIF
#IFDEF __FB_DOS__      ' CP 437
  BoxChar(1) = CHR(201) : BoxChar(2) = CHR(205) : BoxChar(3) = CHR(187)
  BoxChar(4) = CHR(186) : BoxChar(5) = CHR(200) : BoxCHar(6) = CHR(188)
  SymbolChar(1) = CHR(24) : SymbolChar(2) = CHR(176) : SymbolChar(3) = CHR(25)
  DirDelimiter = "\"
#ENDIF
#IFDEF __FB_WIN32__    ' CP 437
  BoxChar(1) = CHR(201) : BoxChar(2) = CHR(205) : BoxChar(3) = CHR(187)
  BoxChar(4) = CHR(186) : BoxChar(5) = CHR(200) : BoxCHar(6) = CHR(188)
  SymbolChar(1) = CHR(24) : SymbolChar(2) = CHR(176) : SymbolChar(3) = CHR(25)
  DirDelimiter = "\"
#ENDIF
  ' 7 bit ASCII:
  'BoxChar(1) = "+" : BoxChar(2) = "-" : BoxChar(3) = "+"
  'BoxChar(4) = "|" : BoxChar(5) = "+" : BoxCHar(6) = "+"
  'SymbolChar(1) = "#" : SymbolChar(2) = "*" : SymbolChar(3) = "#"

#INCLUDE ONCE "wordwrap.bi"   ' WordWrap(BYREF TextToWrap AS STRING, WrapValue AS INTEGER) AS STRING
#INCLUDE ONCE "throwmsg.bi"   ' ThrowMsg(MessageContent AS STRING, SleepTime AS INTEGER = 0)
#INCLUDE ONCE "pophelp.bi"    ' PopHelp(HelpContext AS BYTE)
#INCLUDE ONCE "readtxt.bi"    ' Handler for TXT files
#INCLUDE ONCE "readmenu.bi"   ' Handler for gopher menus
#INCLUDE ONCE "loadfile.bi"   ' File loader
#INCLUDE ONCE "gophurl.bi"    ' Provide the "AnalyseURL()" function

DIM AS INTEGER FirstLoop = 1

CurURL = COMMAND(1)

IF LCASE(CurURL) = "/h" OR LCASE(CurURL) = "-h" OR LCASE(CurURL) = "-help" OR LCASE(CurURL) = "--help" OR CurURL = "/?" THEN
  ShowHelp()
  END(0)
END IF

PRINT "Digger Dwarf v"; pVer; " Copyright (C) Mateusz Viste "; pDate
IF LCASE(LEFT(COMMAND(2), 9)) = "--saveto=" THEN
  SaveTo = TRIM(MID(COMMAND(2), 10))
  IF LEN(SaveTo) = 0 THEN
      PRINT "Please specify a valid filename to the ""--saveto="" parameter."
      END(1)
    ELSE
      IF FileExists(SaveTo) <> 0 THEN
          PRINT "The specified file already exists. Please specify another one."
          END(1)
        ELSE
          IF IsGopherUrlValid(CurURL) = 0 THEN
              PRINT "The specified gopher URL is invalid."
              END(1)
            ELSE
              LoadFile(AnalyseURL(CurURL, "server"), VAL(AnalyseURL(CurURL, "port")), AnalyseURL(CurURL, "selector"), 0, "nowindow")
              IF DumpLoadedFileTo(SaveTo) = 0 THEN
                  PRINT "The file has been saved successfully."
                  END(0)
                ELSE
                  PRINT "An error condition occurred."
                  END(1)
              END IF
          END IF
      END IF
  END IF
END IF

SLEEP 250
OrgCursorPosX = POS(0)
OrgCursorPosY = CSRLIN

GetScreenSize
IF xSize < 40 AND ySize < 20 THEN PRINT "The terminal is too small. The minimal required size is 40 columns x 20 rows."

#IFDEF __FB_DOS__ ' Double-buffering possible only on the DOS version.
  SCREEN , 1, 0  ' Setting page 1 as "active", and page 0 as "display"
  PCOPY 0, 2     ' Page 0: Displayed page ; Page 1: Work page ; Page 2: Saved initial page
#ENDIF

COLOR 7, 0
CLS
LoadSettings()
ClearContentSpace()
DrawTUI()

WHILE IsGopherUrlValid(CurURL) = 0  ' Wrong URL, specify one!
  IF FirstLoop = 0 THEN ThrowMsg("Wrong URL! Please specify a valid URL and try again.", 1500)
  FirstLoop = 0
  ClearContentSpace()
  CurURL = TypeNewURL(CurURL)
WEND

DO
  NewUrl = AnalyseURL(CurURL, "server")
  IF VAL(AnalyseURL(CurURL, "port")) <> 70 THEN NewURL += ":" + AnalyseURL(CurURL, "port")
  NewUrl +=  "/" & AnalyseURL(CurURL, "GopherType") + AnalyseURL(CurURL, "selector")
  IF IsGopherUrlValid(NewURL) <> 0 THEN CurURL = NewURL
  NewURL = ""

  RefreshScreen()
  IF LCASE(LEFT(AnalyseURL(CurURL, "Selector"), 4)) = "url:" OR LCASE(LEFT(AnalyseURL(CurURL, "Selector"), 5)) = "/url:" THEN
      ThrowMsg("You clicked on a non-gopher URL link. Digger Dwarf do not support such links." + CHR(10) + "Please use a client proper to the target protocol." + CHR(10) + CHR(10) + "Press any key...")
      SLEEP
      DoNotReload = 1  ' Avoid reloading the current page from the gopher server
    ELSE
      SELECT CASE AnalyseURL(CurURL, "GopherType")
        CASE "0", "1"
          LoadFile(AnalyseURL(CurURL, "server"), VAL(AnalyseURL(CurURL, "port")), AnalyseURL(CurURL, "selector"))
        CASE "7"
            IF LEN(SearchQuery) = 0 THEN SearchQuery = AskForSearchRequest()
            LoadFile(AnalyseURL(CurURL, "server"), VAL(AnalyseURL(CurURL, "port")), AnalyseURL(CurURL, "selector"), 0, "", SearchQuery)
        CASE "2", "8" ' Not supported
          ThrowMsg("Unhandled file type, can't process. Sorry. Press any key...")
          SLEEP
          DoNotReload = 1
        CASE ELSE
          REM Ask where to save the file
          SaveTo = SelectFileToSave()
          IF LEN(SaveTo) = 0 THEN
              ThrowMsg("The downloading has been aborted!", 1000)
              DoNotReload = 1
            ELSE
              IF FileExists(SaveTo) <> 0 THEN
                  ThrowMsg("The specified file already exists! Press any key, and try again.")
                  SLEEP
                  DoNotReload = 1
                ELSE
                  LoadFile(AnalyseURL(CurURL, "server"), VAL(AnalyseURL(CurURL, "port")), AnalyseURL(CurURL, "selector"))
                  IF DumpLoadedFileTo(SaveTo) = 0 THEN
                      ThrowMsg("The file has been downloaded successfully. Press any key...")
                      SLEEP
                    ELSE
                      ThrowMsg("An error condition occurred. The file could not be saved. Press any key...")
                      SLEEP
                  END IF
              END IF
          END IF
      END SELECT
  END IF
  KeybFlush()
  RefreshScreen()

  SELECT CASE AnalyseURL(CurURL, "GopherType")
    CASE "0"       ' Text file
      ReadTXT()
    CASE "1", "7"       ' Gopher menu (or search result)
      ReadMenu()
    CASE ELSE
      KeybFlush()
      IF LEN(UrlHistory(1)) > 0 THEN
          HistoryBack()
        ELSE
          ClearContentSpace()
          NewURL = TypeNewUrl(CurURL)
      END IF
  END SELECT

  WHILE IsGopherUrlValid(NewURL) = 0  ' Wrong URL, specify one!
    IF FirstLoop = 0 THEN ThrowMsg("Wrong URL! Please specify a valid URL and try again.", 1500)
    FirstLoop = 0
    ClearContentSpace()
    NewURL = TypeNewURL(NewURL)
  WEND

  IF UrlHistory(0) = "Do not historize" THEN    ' Do not historize if the new URL
      UrlHistory(0) = ""                        ' is already from the history.
    ELSE
      AddHistory(CurURL, SearchQuery)
      SearchQuery = ""
  END IF
  CurURL = NewURL
LOOP

REM  ***  END OF THE MAIN PROGRAM - SUBS AND FUNCTIONS FOLLOW  ***


FUNCTION CenterString(WorkString AS STRING, xWidth AS INTEGER) AS STRING
  DIM AS STRING Result
  IF LEN(WorkString) >= xWidth THEN
      Result = WorkString
    ELSE
      Result = SPACE((xWidth - LEN(WorkString)) \ 2) + WorkString + SPACE((xWidth - LEN(WorkString)) \ 2)
  END IF
  IF LEN(Result) < xWidth THEN Result += " "
  RETURN RESULT
END FUNCTION


SUB GetScreenSize
  DIM w AS INTEGER = Width
  xSize = LOWORD(w)
  ySize = HIWORD(w)
END SUB


SUB LoadSettings()
  REM  *** WRITE ME ***
  REM  >> Default Scheme <<
  ColorScheme.TitleBarFg = 14
  ColorScheme.TitleBarBg = 6
  ColorScheme.StatusBarFg = 0
  ColorScheme.StatusBarBg = 3
  ColorScheme.UrlBarFg = 7
  ColorScheme.UrlBarBrightFg = 15
  ColorScheme.UrlBarBg = 9
  ColorScheme.BorderFg = 3
  ColorScheme.BorderBg = 0
  ColorScheme.MsgWindowFg = 14
  ColorScheme.MsgWindowBg = 4
  ColorScheme.MsgWindowBorder = 0
  ColorScheme.DocumentFg = 7
  ColorScheme.DocumentBg = 1
  ColorScheme.GopherSelectedLinkFg = 15
  ColorScheme.GopherSelectedLinkBg = 2
  ColorScheme.GopherItemType = 8
  ColorScheme.GopherSelectedItemType = 0
  ColorScheme.GopherLink = 7
  ColorScheme.GopherBackground = 1
  ColorScheme.GopherInfo = 2
  ColorScheme.GopherError = 4
  ColorScheme.SaveFileDialogPathFg = 15
  ColorScheme.SaveFileDialogPathBg = 12
  ColorScheme.SearchTermDialogFg = 15
  ColorScheme.SearchTermDialogBg = 12
  REM  >> B/W Scheme <<
  'ColorScheme.TitleBarFg = 0
  'ColorScheme.TitleBarBg = 15
  'ColorScheme.StatusBarFg = 8
  'ColorScheme.StatusBarBg = 7
  'ColorScheme.UrlBarFg = 8
  'ColorScheme.UrlBarBrightFg = 0
  'ColorScheme.UrlBarBg = 7
  'ColorScheme.BorderFg = 7
  'ColorScheme.BorderBg = 0
  'ColorScheme.MsgWindowFg = 15
  'ColorScheme.MsgWindowBg = 7
  'ColorScheme.MsgWindowBorder = 0
  'ColorScheme.DocumentFg = 7
  'ColorScheme.DocumentBg = 0
  'ColorScheme.GopherSelectedLinkFg = 0
  'ColorScheme.GopherSelectedLinkBg = 7
  'ColorScheme.GopherItemType = 8
  'ColorScheme.GopherSelectedItemType = 8
  'ColorScheme.GopherLink = 7
  'ColorScheme.GopherBackground = 0
  'ColorScheme.GopherInfo = 7
  'ColorScheme.GopherError = 7
  'ColorScheme.SaveFileDialogPathFg = 15
  'ColorScheme.SaveFileDialogPathBg = 0
  'ColorScheme.SearchTermDialogFg = 15
  'ColorScheme.SearchTermDialogBg = 0
END SUB


SUB ShowHelp()
  PRINT "Digger Dwarf v"; pVer; " Copyright (C) Mateusz Viste "; pDate
  PRINT
  PRINT "Digger Dwarf is a simple console-based gopher client. If you don't know what"
  PRINT "gopher is, then this software will be probably useless to you."
  PRINT
  PRINT "Usage:"
  PRINT "ddwarf [url [--saveto=file]]"
  PRINT
  PRINT "If you call DDwarf without any parameter, then it will simply open its browsing"
  PRINT "window, and let you enter any URL of your choice. If called with a URL"
  PRINT "parameter, it will start and immediately open the specified URL (if possible)."
  PRINT "If you add the ""--saveto="" parameter, then DDwarf will load the given URL, dump"
  PRINT "its content to the specified file, and quit."
  PRINT
  PRINT "Examples:"
  PRINT "ddwarf gopher://gopher.viste-family.net"
  PRINT "ddwarf gopher://a-gopher-url/to/a/file --saveto=myfile.txt"
  PRINT
  PRINT "Digger Dwarf is released on the GNU/GPLv3 license. The whole license text is"
  PRINT "shipped along with the program."
  PRINT
END SUB


SUB RefreshScreen()
  #IFDEF __FB_DOS__  ' Double buffering is active
    PCOPY 1, 0       ' only on the DOS version.
  #ENDIF
END SUB


SUB AddHistory(UrlToAdd AS STRING, SearchQueryToAdd AS STRING = "")
  DIM AS INTEGER x
  DIM AS STRING NewHistoryEntry
  NewHistoryEntry = UrlToAdd
  IF LEN(SearchQueryToAdd) > 0 THEN NewHistoryEntry += CHR(9) + SearchQueryToAdd
  IF NOT NewHistoryEntry = UrlHistory(1) THEN  ' avoid historizing exactly the same ressource twice in a row
    FOR x = UBOUND(UrlHistory) TO 2 STEP -1
      UrlHistory(x) = UrlHistory(x - 1)
    NEXT x
    UrlHistory(1) = NewHistoryEntry
  END IF
END SUB


SUB HistoryBack()
  DIM x AS INTEGER
  IF LEN(UrlHistory(1)) > 0 THEN
      NewURL = UrlHistory(1)
      UrlHistory(0) = "Do not historize"
      FOR x = 1 TO UBOUND(UrlHistory) - 1
        UrlHistory(x) = UrlHistory(x + 1)
      NEXT x
      UrlHistory(UBOUND(UrlHistory)) = ""
      IF INSTR(NewURL, CHR(9)) > 0 THEN    ' If search query found, extract it
          SearchQuery = MID(NewURL, INSTR(NewURL, CHR(9)) + 1)
          NewURL = MID(NewURL, 1, INSTR(NewURL, CHR(9)) - 1)
        ELSE
          SearchQuery = ""
      END IF
    ELSE
      ThrowMsg("No history available, yet.", 1200)
      UrlHistory(0) = "Do not historize"
      NewURL = CurURL
      DoNotReload = 1
  END IF
END SUB


SUB QuitProgram(QuitMessage AS STRING = "")
  COLOR 7, 0
  #IFDEF __FB_DOS__
    PCOPY 2, 0
    LOCATE OrgCursorPosY, OrgCursorPosX
  #ELSE
    CLS
  #ENDIF
  IF LEN(QuitMessage) > 0 THEN PRINT QuitMessage
'  DIM AS INTEGER x        '
'  PRINT                   ' Dump URL history
'  COLOR 7, 0              ' history at exit
'  PRINT "History:"        ' (usefull for debug)
'  FOR x = 0 TO 255        '
'    IF LEN(UrlHistory(x)) > 0 THEN PRINT x & " "; UrlHistory(x)
'  NEXT x
  END
END SUB


SUB KeybFlush()   ' Flush keyboard buffer
  WHILE LEN(INKEY) > 0 : WEND
END SUB


FUNCTION TypeNewURL(ActualUrl AS STRING) AS STRING
  DIM AS STRING InputBuffer, LastKey, Result
  DIM AS INTEGER MaxUrlLen
  MaxUrlLen = xSize - 15
  InputBuffer = ActualUrl  ' Do not put CurURL here, would break things when typing bad new URL...
  NewURL = ""
  DO
    LOCATE 2, 13, 1    ' Locate the cursor + make sure the blinking cursor is visible
    IF LEN(InputBuffer) > MaxUrlLen THEN
        COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
        PRINT "...";
        COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg
        PRINT RIGHT(InputBuffer, MaxUrlLen - 3);
      ELSE
        COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg
        PRINT InputBuffer;
        COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
        PRINT RepeatPrint(".", MaxUrlLen - LEN(InputBuffer));
    END IF
    LOCATE 2, 13 + LEN(InputBuffer)
    KeybFlush()
    SLEEP
    LastKey = INKEY
    IF ASC(LastKey) > 31 AND LEN(LastKey) = 1 THEN InputBuffer += LastKey   ' Ignore any control codes
    IF ASC(LastKey) = 8 AND LEN(InputBuffer) > 0 THEN InputBuffer = LEFT(InputBuffer, LEN(InputBuffer) - 1)
    IF ASC(LastKey) = 9 THEN Result = CurURL: UrlHistory(0) = "Do not historize": DoNotReload = 1
    IF ASC(LastKey) = 27 THEN QuitProgram()
    IF LastKey = CHR(13) THEN Result = InputBuffer
  LOOP UNTIL LEN(Result) > 0 ' OR LastKey = CHR(13)
  RETURN Result
END FUNCTION


FUNCTION RepeatPrint(RepeatString AS STRING, RepeatAction AS INTEGER) AS STRING
  DIM AS INTEGER x
  DIM AS STRING Result = ""
  FOR x = 1 TO RepeatAction
    Result += RepeatString
  NEXT x
  RETURN Result
END FUNCTION


SUB ClearContentSpace()
  DIM AS INTEGER x
  COLOR 0, ColorScheme.GopherBackground
  FOR x = 3 TO ySize - 1
    LOCATE x, 2 : PRINT SPACE(xSize - 2);
  NEXT x
END SUB


SUB DrawTUI()
  DIM AS INTEGER y, MaxUrlLen
  MaxUrlLen = xSize - 15
  LOCATE 1, 1: COLOR ColorScheme.TitleBarFg, ColorScheme.TitleBarBg : PRINT CenterString("Digger Dwarf v" + pVer, xSize);
  COLOR ColorScheme.BorderFg, ColorScheme.BorderBg
  LOCATE 2, 1 : PRINT BoxChar(1);
  LOCATE 2, xSize : PRINT BoxChar(3);
  FOR y = 3 TO ySize - 1
    LOCATE y, 1: PRINT BoxChar(4);
    LOCATE y, xSize: PRINT BoxChar(4);
  NEXT y
  COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg
  LOCATE 2, 2: COLOR ColorScheme.UrlBarFg, ColorScheme.UrlBarBg : PRINT " [" + STRING(xSize - 6, ASC(".")) + "] ";
  LOCATE 2, 4: COLOR ColorScheme.UrlBarBrightFg, ColorScheme.UrlBarBg: PRINT "gopher://";
  IF LEN(CurURL) > MaxUrlLen THEN
      PRINT LEFT(CurURL, MaxUrlLen - 5); "(...)";
    ELSE
      PRINT CurURL;
  END IF
  COLOR ColorScheme.StatusBarFg, ColorScheme.StatusBarBg
  LOCATE ySize, 1: PRINT SPACE(xSize);
END SUB


FUNCTION SelectFileToSave() AS STRING
  DIM AS STRING Result
  DIM AS INTEGER xStart, yStart
  xStart = (xSize \ 2) - 20 + 1
  yStart = (ySize \ 2) - 5
  COLOR ColorScheme.MsgWindowBorder, ColorScheme.MsgWindowBg
  LOCATE yStart, xStart: PRINT BoxChar(1); RepeatPrint(BoxChar(2), 9); "[ Download a file ]"; RepeatPrint(BoxChar(2), 10); BoxChar(3);
  LOCATE yStart + 1, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 2, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 3, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 4, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 5, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 6, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 7, xStart: PRINT BoxChar(5); RepeatPrint(BoxChar(2), 38); BoxChar(6);
  COLOR ColorScheme.MsgWindowFg, ColorScheme.MsgWindowBg
  LOCATE yStart + 1, xStart + 2: PRINT "The selected file type is not       ";
  LOCATE yStart + 2, xStart + 2: PRINT "supported natively by Digger Dwarf. ";
  LOCATE yStart + 3, xStart + 2: PRINT "Please enter the filename to which  ";
  LOCATE yStart + 4, xStart + 2: PRINT "the file will be downloaded.        ";
  LOCATE yStart + 5, xStart + 2: PRINT "(Press the <ESCAPE> key to abort)   ";
  Result = CURDIR ' pre-fill the download path to current directory
  IF RIGHT(Result, 1) <> DirDelimiter THEN Result += DirDelimiter
  Result = GetUserString(xStart + 2, yStart + 6, 36, ColorScheme.SaveFileDialogPathFg, ColorScheme.SaveFileDialogPathBg, Result)
  RETURN Result
END FUNCTION


FUNCTION DumpLoadedFileTo(FileToDumpTo AS STRING) AS BYTE
  REM  Dumps the loaded file to a local file. Returns 0 if successfull, -1 otherwise.
  DIM AS INTEGER x, SaveFileHandler
  DIM AS BYTE Result
  SaveFileHandler = FREEFILE
  IF FileSize > UBOUND(FileBuffer) THEN PRINT "DEBUG: Error condition, FileSize > FileBuffer!" : END(1)
  Result = OPEN(FileToDumpTo, FOR OUTPUT, AS #SaveFileHandler)
  IF Result = 0 THEN
      FOR x = 1 TO FileSize
        PRINT #SaveFileHandler, CHR(FileBuffer(x));
      NEXT x
      CLOSE #SaveFileHandler
    ELSE
      Result = -1
  END IF
  RETURN Result
END FUNCTION


FUNCTION GetUserString(xposition AS INTEGER, yposition AS INTEGER, fieldwidth AS INTEGER, FgColor AS INTEGER, BgColor AS INTEGER, PreFilledString AS STRING) AS STRING
  DIM AS STRING Result, WorkString, LastKey
  Result = PreFilledString
  WHILE LastKey <> CHR(13) AND LastKey <> CHR(27)
    LOCATE yposition, xposition, 1    ' Locate the cursor + make sure that it's blinking
    COLOR FgColor, BgColor: PRINT SPACE(fieldwidth);
    WorkString = Result
    IF LEN(Result) > fieldwidth THEN WorkString = "..." + RIGHT(WorkString, fieldwidth - 3)
    LOCATE yposition, xposition : PRINT WorkString;
    LastKey = ""
    KeybFlush()
    DO
      SLEEP 1,1
      LastKey = INKEY
    LOOP UNTIL LEN(LastKey) > 0
    IF LastKey = CHR(27) THEN Result = ""
    IF LEN(LastKey) = 1 AND ASC(LastKey) >= 32 THEN Result += LastKey
    IF LastKey = CHR(8) AND LEN(Result) > 0 THEN Result = LEFT(Result, LEN(Result) - 1)
  WEND
  RETURN Result
END FUNCTION


FUNCTION AskForSearchRequest() AS STRING
  DIM AS STRING Result
  DIM AS INTEGER xStart, yStart
  xStart = (xSize \ 2) - 20 + 1
  yStart = (ySize \ 2) - 5
  COLOR ColorScheme.MsgWindowBorder, ColorScheme.MsgWindowBg
  LOCATE yStart, xStart: PRINT BoxChar(1); RepeatPrint(BoxChar(2), 38); BoxChar(3);
  LOCATE yStart + 1, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 2, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 3, xStart: PRINT BoxChar(4); RepeatPrint(" ", 38); BoxChar(4);
  LOCATE yStart + 4, xStart: PRINT BoxChar(5); RepeatPrint(BoxChar(2), 38); BoxChar(6);
  COLOR ColorScheme.MsgWindowFg, ColorScheme.MsgWindowBg
  LOCATE yStart + 1, xStart + 2: PRINT "Enter a query...                    ";
  Result = GetUserString(xStart + 2, yStart + 3, 36, ColorScheme.SearchTermDialogFg, ColorScheme.SearchTermDialogBg, "")
  RETURN Result
END FUNCTION


SUB SetupMenu()
  ThrowMsg("Setup menu not yet implemented! Sorry.", 2000)
END SUB


SUB UpdateStatusBar(StatusInformation AS STRING)
  DIM AS STRING StatusBar
  COLOR ColorScheme.StatusBarFg, ColorScheme.StatusBarBg
  LOCATE ySize, 1: PRINT SPACE(xSize);
  IF LEN(StatusInformation) > xSize - 16 THEN
      StatusBar = LEFT(StatusInformation, xSize - 19) + "..." + " (F1 for Help)"
    ELSE
      StatusBar = StatusInformation + SPACE(xSize - LEN(StatusInformation) - 16) + " (F1 for Help)"
  END IF
  LOCATE ySize, 1: PRINT " " + StatusBar + " ";
END SUB