REM
REM  DDwarf handler for gopher menus
REM
REM  ReadMenu is expecting the following shared variables to be set:
REM    xSize.........................Terminal's max columns
REM    ySize.........................Terminsl's max rows
REM    FileBuffer()..................Array containing the data file
REM    FileSize......................Data file's size in bytes
REM    ColorScheme.GopherInfo........Content window foreground
REM    ColorScheme.GopherError.......Content window foreground
REM    ColorScheme.GopherLink........Content window foreground
REM    ColorScheme.GopherBackground..Content window foreground
REM    ColorScheme.StatusBarFg.......Status bar foreground
REM    ColorScheme.StatusBarBg.......Status bar background
REM    ColorScheme.BorderFg..........Border's foreground color
REM    ColorScheme.BorderBg..........Border's background color
REM
#INCLUDE ONCE "wordwrap.bi"
#INCLUDE ONCE "vbcompat.bi"

PRIVATE FUNCTION ExplodeGopherItem(RawLine AS STRING, nRes AS BYTE) AS STRING
  DIM Result(0 TO 5) AS STRING
  DIM AS BYTE MalformedItem, FieldCounter
  DIM AS INTEGER x
  IF LEN(RawLine) > 1 THEN
        Result(1) = LEFT(RawLine, 1)
      ELSE
        MalformedItem = 1
    END IF
    x = 2
    FieldCounter = 2
    WHILE x <= LEN(RawLine) AND FieldCounter <= 5
      IF RawLine[x - 1] <> 9 THEN Result(FieldCounter) += CHR(RawLine[x - 1]) ELSE FieldCounter += 1
      x += 1
    WEND
    IF LEN(Result(1)) = 0 THEN MalformedItem = 1
    IF LEN(Result(2)) = 0 AND Result(1) <> "i" THEN MalformedItem = 1
    IF MalformedItem = 1 THEN
      Result(1) = "3"
      Result(2) = "(The server sent a malformed item here)"
      Result(3) = ""
      Result(4) = ""
      Result(5) = "0"
    END IF
    IF nRes > 0 AND nRes < 6 THEN Result(0) = Result(nRes) ELSE Result(0) = ""
  RETURN Result(0)
END FUNCTION


PRIVATE FUNCTION GopherTypesDesc(RawType AS STRING) AS STRING
  DIM AS STRING Result
  SELECT CASE RawType
    CASE "i", "3"
      Result = "[INF]"
    CASE "0"
      Result = "[TXT]"
    CASE "1"
      Result = "[DIR]"
    CASE "2"
      Result = "[CSO]"
    CASE "5", "9"
      Result = "[BIN]"
    CASE "7"
      Result = "[QRY]"
    CASE "8"
      Result = "[TEL]"
    CASE "g"
      Result = "[GIF]"
    CASE "I", ":"
      Result = "[IMG]"
    CASE "<", "s"
      Result = "[SND]"
    CASE "P"
      Result = "[PDF]"
    CASE "h"
      Result = "[HTM]"
    CASE ELSE
      Result = "[???]"
  END SELECT
  RETURN Result
END FUNCTION


SUB ReadMenu()
  DIM AS STRING LineBuff, LastKey
  DIM GopherItem(1 TO 5) AS STRING
  DIM AS UBYTE CharBuff
  REDIM ItemsType(1 TO 16) AS STRING
  REDIM ItemsTitle(1 TO 16) AS STRING
  REDIM ItemsSelector(1 TO 16) AS STRING
  REDIM ItemsServer(1 TO 16) AS STRING
  REDIM ItemsPort(1 TO 16) AS INTEGER
  DIM AS INTEGER FileCursor, MaxLineLen, FirstLine = 1, x, MaxLineCount, CursorPosition, ItemsCount, SelectedItem = 0, FirstSelectableItem = 0, LastSelectableItem = 0, NextSelectableItem
  DIM AS DOUBLE LastRefresh = TIMER
  REDIM ScreenBuffer(1 TO 32) AS STRING
  MaxLineLen = xSize - 4
  MaxLineCount = ySize - 3
  ItemsCount = 1        ' Adding an empty
  ItemsType(1) = "i"    ' entry to the top
  DrawTUI()
  WHILE FileCursor < FileSize
    LineBuff = ""
    DO
      FileCursor += 1
      CharBuff = FileBuffer(FileCursor)
      IF CharBuff >= 32 OR CharBuff = 9 THEN LineBuff += CHR(CharBuff)  ' Ignore control codes
    LOOP UNTIL FileCursor >= FileSize OR CharBuff = 10
    IF NOT (LineBuff = "." AND FileCursor = FileSize) THEN ' Ignore a single dot at the end of the transmission
      IF ItemsCount = UBOUND(ItemsType) THEN                         '
        REDIM PRESERVE ItemsType(1 TO ItemsCount * 2) AS STRING      ' Add an empty
        REDIM PRESERVE ItemsTitle(1 TO ItemsCount * 2) AS STRING     ' item to the
        REDIM PRESERVE ItemsSelector(1 TO ItemsCount * 2) AS STRING  ' items list.
        REDIM PRESERVE ItemsServer(1 TO ItemsCount * 2) AS STRING    '
        REDIM PRESERVE ItemsPort(1 TO ItemsCount * 2) AS INTEGER     '
      END IF                                                         '
      ItemsCount += 1                                                '
      ItemsType(ItemsCount) = ExplodeGopherItem(LineBuff, 1)      ' Fill in all
      ItemsTitle(ItemsCount) = ExplodeGopherItem(LineBuff, 2)     ' gopher properties
      ItemsSelector(ItemsCount) = ExplodeGopherItem(LineBuff, 3)  ' for the given
      ItemsServer(ItemsCount) = ExplodeGopherItem(LineBuff, 4)    ' item.
      ItemsPort(ItemsCount) = VAL(ExplodeGopherItem(LineBuff, 5)) '
      IF ItemsType(ItemsCount) <> "i" AND ItemsType(ItemsCount) <> "3" THEN
        IF UCASE(LEFT(ItemsSelector(ItemsCount), 4)) = "URL:" OR UCASE(LEFT(ItemsSelector(ItemsCount), 5)) = "/URL:" THEN
            ItemsTitle(ItemsCount) = "[URL] " + ItemsTitle(ItemsCount)
          ELSE
            ItemsTitle(ItemsCount) = GopherTypesDesc(ItemsType(ItemsCount)) + " " + ItemsTitle(ItemsCount)
        END IF
        IF SelectedItem = 0 THEN SelectedItem = ItemsCount
        IF FirstSelectableItem = 0 THEN FirstSelectableItem = ItemsCount
        LastSelectableItem = ItemsCount
      END IF
      IF ItemsPort(ItemsCount) = 0 THEN ItemsPort(ItemsCount) = 70
      IF LEN(ItemsTitle(ItemsCount)) > MaxLineLen THEN
        IF ItemsType(ItemsCount) = "i" OR ItemsType(ItemsCount) = "3" THEN
            LineBuff = ItemsTitle(ItemsCount)
            ItemsTitle(ItemsCount) = WordWrap(LineBuff, MaxLineLen)
            WHILE LEN(LineBuff) > 0
              IF ItemsCount = UBOUND(ItemsType) THEN                         '
                REDIM PRESERVE ItemsType(1 TO ItemsCount * 2) AS STRING      ' Add an empty
                REDIM PRESERVE ItemsTitle(1 TO ItemsCount * 2) AS STRING     ' item to the
                REDIM PRESERVE ItemsSelector(1 TO ItemsCount * 2) AS STRING  ' items list.
                REDIM PRESERVE ItemsServer(1 TO ItemsCount * 2) AS STRING    '
                REDIM PRESERVE ItemsPort(1 TO ItemsCount * 2) AS INTEGER     '
              END IF                                                         '
              ItemsCount += 1                                                '
              ItemsType(ItemsCount) = ItemsType(ItemsCount - 1)
              ItemsTitle(ItemsCount) = WordWrap(LineBuff, MaxLineLen)
              ItemsSelector(ItemsCount) = ItemsSelector(ItemsCount - 1)
              ItemsServer(ItemsCount) = ItemsServer(ItemsCount - 1)
              ItemsPort(ItemsCount) = ItemsPort(ItemsCount - 1)
            WEND
          ELSE
            ItemsTitle(ItemsCount) = LEFT(ItemsTitle(ItemsCount), MaxLineLen - 3) + "..."
        END IF
      END IF
    END IF
  WEND
  REM  -==  Okay, all gopher items have been processed. Let's display them now!  ==-
  DO
    FOR x = 0 TO MaxLineCount - 1
      LOCATE 3 + x, 2, 0  ' Locate the cursor and make sure that the blinking crap is off
      IF FirstLine + x <= ItemsCount THEN
          SELECT CASE ItemsType(FirstLine + x)
            CASE "i"
              COLOR ColorScheme.GopherInfo, ColorScheme.GopherBackground
              PRINT " " + ItemsTitle(FirstLine + x);
            CASE "3"
              COLOR ColorScheme.GopherError, ColorScheme.GopherBackground
              PRINT " " + ItemsTitle(FirstLine + x);
            CASE ELSE
              IF FirstLine + x = SelectedItem THEN
                  COLOR ColorScheme.GopherSelectedItemType, ColorScheme.GopherSelectedLinkBg
                ELSE
                  COLOR ColorScheme.GopherItemType, ColorScheme.GopherBackground
              END IF
              PRINT " " + LEFT(ItemsTitle(FirstLine + x), 5);
              IF FirstLine + x = SelectedItem THEN
                  COLOR ColorScheme.GopherSelectedLinkFg, ColorScheme.GopherSelectedLinkBg
                ELSE
                  COLOR ColorScheme.GopherLink, ColorScheme.GopherBackground
              END IF
              PRINT MID(ItemsTitle(FirstLine + x), 6);
          END SELECT
          REM Fill the rest of the line with blanks. GIVES CRAP RESULTS ON UTF-8 (wrong length calculation)!
          PRINT SPACE(MaxLineLen - LEN(ItemsTitle(FirstLine + x))) + " ";
        ELSE
          COLOR ColorScheme.GopherInfo, ColorScheme.GopherBackground
          PRINT SPACE(MaxLineLen + 2);
      END IF
    NEXT x
    IF ItemsCount > MaxLineCount THEN    ' Refresh the scrollbar
      COLOR ColorScheme.BorderFg, ColorScheme.BorderBg
      CursorPosition = INT((FirstLine - 1) / (ItemsCount - MaxLineCount) * (MaxLineCount - 3))
      LOCATE 3, xSize : PRINT SymbolChar(1);
      LOCATE ySize - 1, xSize : PRINT SymbolChar(3);
      FOR x = 4 TO ySize - 2
        LOCATE x, xSize : PRINT SymbolChar(2);
      NEXT x
      COLOR ColorScheme.BorderBg, ColorScheme.BorderFg
      LOCATE 4 + CursorPosition, xSize : PRINT " ";
    END IF
    REM Update the status bar to reflect the current selected item (if any item is selected)
    IF SelectedItem > 0 THEN
        IF UCASE(LEFT(ItemsSelector(SelectedItem), 4)) = "URL:" THEN
            UpdateStatusBar("URL: " & MID(ItemsSelector(SelectedItem), 5))
          ELSE
            IF ItemsPort(SelectedItem) = 70 THEN
                UpdateStatusBar("URL: gopher://" & ItemsServer(SelectedItem) & "/" & ItemsType(SelectedItem) & ItemsSelector(SelectedItem))
              ELSE
                UpdateStatusBar("URL: gopher://" & ItemsServer(SelectedItem) & ":" & ItemsPort(SelectedItem) & "/" & ItemsType(SelectedItem) & ItemsSelector(SelectedItem))
            END IF
        END IF
      ELSE
        UpdateStatusBar("No selectable item found.")
    END IF
    REM Status bar updated
    RefreshScreen()  ' Refresh screen content
    KeybFlush()      ' Flush keyboard buffer
    SLEEP
    LastKey = INKEY
    SELECT CASE LastKey
      CASE CHR(255) + "H"   ' Up
        IF SelectedItem > 0 AND SelectedItem > FirstSelectableItem THEN
            DO
              IF SelectedItem < FirstLine THEN
                  FirstLine -= 1
                ELSE
                NextSelectableItem = SelectedItem
                DO
                  NextSelectableItem -= 1
                LOOP UNTIL ItemsType(NextSelectableItem) <> "i" AND ItemsType(NextSelectableItem) <> "3"
                SELECT CASE NextSelectableItem
                  CASE IS = FirstLine - 1
                    SelectedItem = NextSelectableItem
                    FirstLine -= 1
                  CASE IS < FirstLine - 1
                    FirstLine -= 1
                  CASE ELSE
                    SelectedItem = NextSelectableItem
                END SELECT
              END IF
            LOOP UNTIL ItemsType(SelectedItem) <> "i" AND ItemsType(SelectedItem) <> "3"
          ELSEIF (SelectedItem = FirstSelectableItem OR SelectedItem = 0) AND FirstLine > 1 THEN
            FirstLine -= 1
        END IF
      CASE CHR(255) + "P"   ' Down
        IF SelectedItem > 0 AND SelectedItem < LastSelectableItem THEN
            IF SelectedItem > FirstLine + MaxLineCount - 1 THEN
                FirstLine += 1
              ELSE
                NextSelectableItem = SelectedItem
                DO
                  NextSelectableItem += 1
                LOOP UNTIL ItemsType(NextSelectableItem) <> "i" AND ItemsType(NextSelectableItem) <> "3"
                SELECT CASE NextSelectableItem
                  CASE IS = FirstLine + MaxLineCount
                    SelectedItem = NextSelectableItem
                    FirstLine += 1
                  CASE IS > FirstLine + MaxLineCount
                    FirstLine += 1
                  CASE ELSE
                    SelectedItem = NextSelectableItem
                END SELECT
            END IF
          ELSEIF FirstLine < ItemsCount - MaxLineCount + 1 THEN
            FirstLine += 1
        END IF
      CASE CHR(255, 71)     ' Home
        FirstLine = 1
        SelectedItem = FirstSelectableItem
      CASE CHR(255, 79)     ' End
        FirstLine = ItemsCount - MaxLineCount + 1
        IF FirstLine < 1 THEN FirstLine = 1
        SelectedItem = LastSelectableItem
      CASE CHR(13)          ' Enter
        IF SelectedItem > 0 THEN
          NewURL = ItemsServer(SelectedItem) + ":" & ItemsPort(SelectedItem) & "/" + ItemsType(SelectedItem) + ItemsSelector(SelectedItem)
          IF IsGopherUrlValid(NewURL) = 0 THEN
            NewURL = CurURL
            ThrowMsg("The selected link appears to be broken. I won't follow it!" + CHR(10) + CHR(10) + "Press any key...", -1)
          END IF
        END IF
      CASE CHR(8)           ' BackSpace
        HistoryBack()
      CASE CHR(9)           ' TAB
        NewURL = TypeNewURL(CurURL)
      CASE CHR(27)          ' Escape
        QuitProgram()
      CASE CHR(255, 59)     ' F1
        PopHelp(1)
      CASE CHR(255, 63)     ' F5
        IF ABS(TIMER - LastRefresh) >= 0.5 THEN ' Do not allow refreshing
            UrlHistory(0) = "Do not historize"  ' quicker than twice per
            NewURL = CurURL                     ' second.
          ELSE
            ThrowMsg("Refreshing rate is limited to 2 times a second maximum. Calm down, speedy!", 4000)
        END IF
      CASE CHR(255, 68)     ' F10
        SetupMenu()
    END SELECT
    IF SelectedItem > 0 THEN
      IF ItemsType(SelectedItem) = "i" OR ItemsType(SelectedItem) = "3" AND SelectedItem > 0 THEN
        DO: SelectedItem += 1
          IF SelectedItem > ItemsCount THEN SelectedItem = 1 : FirstLine = 1
        LOOP UNTIL (ItemsType(SelectedItem) <> "i" AND ItemsType(SelectedItem) <> "3")
      END IF
    END IF
  LOOP UNTIL LEN(NewURL) > 0
END SUB
