REM
REM  FDGOPHER handler for TXT files
REM
REM  ReadTXT is expecting the following shared variables to be set:
REM    xSize....................Terminal's max columns
REM    ySize....................Terminsl's max rows
REM    FileBuffer().............Array containing the data file
REM    FileSize.................Data file's size in bytes
REM    ColorScheme.DocumentFg...Content window foreground
REM    ColorScheme.DocumentBg...Content window background
REM    ColorScheme.StatusBarFg..Status bar foreground
REM    ColorScheme.StatusBarBg..Status bar background
REM    ColorScheme.BorderFg.....Border's foreground color
REM    ColorScheme.BorderBg.....Border's background color
REM
#INCLUDE ONCE "wordwrap.bi"
#INCLUDE ONCE "vbcompat.bi"

SUB ReadTxt()
  DIM AS STRING LineBuff, LastKey, WrappedLine
  DIM AS UBYTE CharBuff
  DIM AS INTEGER FileCursor, ScreenBufferLines, MaxLineLen, FirstLine = 1, x, MaxLineCount, CursorPosition
  DIM AS DOUBLE LastRefresh = TIMER
  REDIM ScreenBuffer(1 TO 32) AS STRING
  MaxLineLen = xSize - 2
  MaxLineCount = ySize - 3
  DrawTUI()
  WHILE FileCursor < FileSize
    DO
      FileCursor += 1
      CharBuff = FileBuffer(FileCursor)
      IF CharBuff >= 32 THEN LineBuff += CHR(CharBuff)  ' Ignore control codes
      IF CharBuff = 9 THEN LineBuff += SPACE(8)      ' Replace any TAB by 8 spaces
    LOOP UNTIL FileCursor >= FileSize OR CharBuff = 10
    IF NOT (LineBuff = "." AND FileSize - FileCursor < 3) THEN
      DO
        WrappedLine = WordWrap(LineBuff, MaxLineLen)
        IF ScreenBufferLines = UBOUND(ScreenBuffer) THEN REDIM PRESERVE ScreenBuffer(1 TO ScreenBufferLines * 2) AS STRING
        ScreenBufferLines += 1
        ScreenBuffer(ScreenBufferLines) = WrappedLine
      LOOP UNTIL LineBuff = ""
    END IF
    LineBuff = ""
  WEND
  UpdateStatusBar("File size: " & FORMAT(FileSize / 1024, "#0.00") & " KiB")

  DO
    COLOR ColorScheme.DocumentFg, ColorScheme.DocumentBg
    FOR x = 1 TO MaxLineCount
      IF FirstLine + x - 1 <= ScreenBufferLines THEN LineBuff = ScreenBuffer(FirstLine + x - 1) ELSE LineBuff = ""
      LOCATE 2 + x, 2: PRINT SPACE(MaxLineLen); ' You could as well simply display the line of content, and
      LOCATE 2 + x, 2: PRINT LineBuff;          ' fill lacking whitespaces, but it gives crappy results on UTF8...
    NEXT x
    IF ScreenBufferLines > MaxLineCount THEN    ' Refresh the scrollbar
      COLOR ColorScheme.BorderFg, ColorScheme.BorderBg
      CursorPosition = INT((FirstLine - 1) / (ScreenBufferLines - MaxLineCount) * (MaxLineCount - 3))
      LOCATE 3, xSize : PRINT SymbolChar(1);
      LOCATE ySize - 1, xSize : PRINT SymbolChar(3);
      FOR x = 4 TO ySize - 2
        LOCATE x, xSize : PRINT SymbolChar(2);
      NEXT x
      COLOR ColorScheme.BorderBg, ColorScheme.BorderFg
      LOCATE 4 + CursorPosition, xSize : PRINT " ";
    END IF
    LOCATE ySize, xSize, 0   ' Hide the annoying blinking cursor
    RefreshScreen()          ' Update the screen's content
    KeybFlush()              ' Flush keyb buffer
    SLEEP
    LastKey = INKEY
    IF LastKey = CHR(255) + "H" THEN FirstLine -= 1                                   ' Up
    IF LastKey = CHR(255) + "P" THEN FirstLine += 1                                   ' Down
    IF LastKey = CHR(255) + "I" THEN FirstLine -= MaxLineCount                        ' PgUp
    IF LastKey = CHR(255) + "Q" THEN FirstLine += MaxLineCount                        ' PgDown
    IF LastKey = CHR(255) + "G" THEN FirstLine = 1                                    ' Home
    IF LastKey = CHR(255) + "O" THEN FirstLine = ScreenBufferLines - MaxLineCount + 1 ' End
    IF FirstLine + MaxLineCount - 1 > ScreenBufferLines THEN FirstLine = ScreenBufferLines - MaxLineCount + 1
    IF FirstLine > 1 AND ScreenBufferLines <= MaxLineCount THEN FirstLine = 1
    IF FirstLine < 1 THEN FirstLine = 1
    IF Lastkey = CHR(255, 63) THEN         ' F5
      IF ABS(TIMER - LastRefresh) >= 0.5 THEN ' Do not allow refreshing
          UrlHistory(0) = "Do not historize"  ' quicker than twice per
          NewURL = CurURL                     ' second.
        ELSE
          ThrowMsg("Refreshing rate is limited to 2 times a second maximum. Calm down, speedy!", 4000)
      END IF
    END IF
    IF LastKey = CHR(27) THEN QuitProgram()
    IF LastKey = CHR(8) THEN HistoryBack()
    IF LastKey = CHR(9) THEN NewURL = TypeNewURL(CurURL)
    IF LastKey = CHR(255, 59) THEN PopHelp(2)  ' F1
  LOOP UNTIL LEN(NewURL) > 0 'LastKey = CHR(8) OR LastKey = CHR(9)
END SUB
